from pycountry import countries

from phone_iso3166.e164 import mapping


def extract_country_codes(prefix_tree):
    for country_code in prefix_tree.values():
        if isinstance(country_code, dict):
            yield from extract_country_codes(country_code)
        else:
            yield country_code


def test_verify_iso3166():
    for contry_code in extract_country_codes(mapping):
        if contry_code in ("DG", "XK"):
            # Some places have calling prefixes, but are not recognized as
            # countries in their own right. This leads to edgecases.
            continue
        country = countries.get(alpha_2=contry_code)
        assert country is not None
        assert country.alpha_2 == contry_code



class InvalidPhone(Exception):
    """The provided phone number was invalid."""


class InvalidNetwork(Exception):
    """The provided network was invalid."""


class InvalidCountry(Exception):
    """The provided country code was invalid."""



# pylint: disable=consider-using-f-string
from .e164 import mapping
from .e212 import networks
from .errors import InvalidCountry
from .errors import InvalidNetwork
from .errors import InvalidPhone


def phone_country(phone):  # pylint: disable=inconsistent-return-statements
    """Return the ISO-3166-1 country code a given phone number belongs to.

    The phone number must be specified in E.164, aka international format.
    The returned country code is of ISO-3166-1 alpha-2 code.
    """

    e164_mapping = mapping
    try:
        for digit in filter(str.isdigit, str(phone)):
            e164_mapping = e164_mapping[int(digit)]
            if isinstance(e164_mapping, str):
                return e164_mapping
    except Exception as ex:
        raise InvalidPhone("Invalid phone {}".format(phone)) from ex


def country_prefixes():
    """Return a mapping of country code to calling prefix.

    For countries with multiple prefixes, an arbitrary calling prefix is
    chosen. Many north american country codes just map to 1.
    """

    def transverse(node, path):
        if isinstance(node, dict):
            for key, value in node.items():
                for i in transverse(value, path + str(key)):
                    yield i
        else:
            yield node, 1 if node in ["US", "CA"] else int(path)

    return dict(transverse(mapping, ""))


def country_prefix(country_code):
    """Return the calling prefix for a given country.

    The country must be provided as a ISO-3166-1 alpha-2 country code.
    """
    prefixes = country_prefixes()
    try:
        return prefixes[country_code.upper()]
    except Exception as ex:
        raise InvalidCountry("Invalid country {}".format(country_code)) from ex


def phone_country_prefix(phone):  # pylint: disable=inconsistent-return-statements
    """Figure out that calling prefix and country a phone number belongs to.

    The country returned is in the form of a ISO-3166-1 alpha-2 country code.
    """
    e164_mapping = mapping
    prefix = ""
    try:
        for digit in filter(str.isdigit, str(phone)):
            e164_mapping = e164_mapping[int(digit)]
            prefix += digit
            if isinstance(e164_mapping, str):
                return (int(prefix), e164_mapping)
    except Exception as ex:
        raise InvalidPhone("Invalid phone {}".format(phone)) from ex


def network_country(mcc, mnc):
    """Return the country matching mcc and mnc.

    A few countries share mcc, so to work in more cases the mnc needs to be
    provided for a secondary lookup.

    Returns an ISO-3166-1 alpha-2 country code.
    """
    try:
        country = networks[mcc]
        if isinstance(country, str):
            return country
        return country[mnc]
    except Exception as ex:
        raise InvalidNetwork("Invalid MCC {} MNC {}".format(mcc, mnc)) from ex



# pylint: disable=consider-using-f-string
from .e212_names import countries
from .e212_names import operators
from .errors import InvalidCountry
from .errors import InvalidNetwork


def network(mcc, mnc):
    """Return country code and network name for a mcc and mnc combination.

    The country code is returned as a ISO-3166-1 alpha-2 code.
    """
    mcc = int(mcc)
    mnc = int(mnc)
    try:
        return operators[mcc][mnc]
    except Exception as ex:
        raise InvalidNetwork("Invalid MCC {} MNC {}".format(mcc, mnc)) from ex


def country_networks(country):
    """Return tuples of all networks for a given country.

    The country must be provided as a ISO-3166-1 alpha-2 country code.
    The returned for is (mcc, mnc, network_name).
    """
    try:
        return [(m[0], m[1], operators[m[0]][m[1]][1]) for m in countries[country]]
    except Exception as ex:
        raise InvalidCountry("Invalid country {}".format(country)) from ex



import pytest

from phone_iso3166.country import country_prefix
from phone_iso3166.country import country_prefixes
from phone_iso3166.country import network_country
from phone_iso3166.country import phone_country
from phone_iso3166.country import phone_country_prefix
from phone_iso3166.errors import InvalidCountry
from phone_iso3166.errors import InvalidNetwork
from phone_iso3166.errors import InvalidPhone


@pytest.mark.parametrize(
    "phone_number, country_code",
    [
        (45, "DK"),
        (4566118311, "DK"),
        ("+4566118311", "DK"),
        ("+38640118311", "SI"),
        ("+38340118311", "XK"),
        ("+37740118311", "MC"),
        ("+77401183119", "KZ"),
        ("+79901185311", "RU"),
        ("+97040118311", "PS"),
    ],
)
def test_phone_country_dk(phone_number, country_code):
    assert phone_country(phone_number) == country_code


def test_country_prefixes():
    assert isinstance(country_prefixes(), dict)


@pytest.mark.parametrize(
    "country_code, calling_prefix",
    [
        ("dk", 45),
        ("DK", 45),
        ("Dk", 45),
        ("us", 1),
        ("xK", 383),
    ],
)
def test_country_prefix(country_code, calling_prefix):
    assert country_prefix(country_code) == calling_prefix


def test_country_invalid():
    with pytest.raises(InvalidCountry):
        country_prefix("dkk")


def test_invalid():
    with pytest.raises(InvalidPhone):
        phone_country(0)


def test_missing():
    with pytest.raises(InvalidPhone):
        phone_country(999)

    with pytest.raises(InvalidPhone):
        phone_country("999")


@pytest.mark.parametrize(
    "phone_number, country_code",
    [
        ("+1 202-456-1111", "US"),
        (14412921234, "BM"),
    ],
)
def test_country_lookup(phone_number, country_code):
    assert phone_country(phone_number) == country_code


@pytest.mark.parametrize(
    "mcc, mnc, country_code",
    [
        (238, 1, "DK"),  # Denmark
        (340, 1, "GP"),  # Guadeloupe
        (340, 12, "GP"),  # Also Guadeloupe
        (425, 5, "PS"),  # Palestine
        (425, 6, "PS"),  # Palestine
        (425, 7, "IL"),  # Israel
        (505, 20, "AU"),  # Australia
        (362, 31, "BQ"),  # Sint Eustatius, formerly Netherlands Antilles
        (362, 51, "SX"),  # Sint Maarten, formerly Netherlands Antillies
        (647, 0, "RE"),  # Reunion
        (647, 1, "YT"),  # Mayotte
    ],
)
def test_network_multi(mcc, mnc, country_code):
    assert network_country(mcc, mnc) == country_code


def test_network_invalid():
    with pytest.raises(InvalidNetwork):
        network_country(0, 0)


@pytest.mark.parametrize(
    "phone_number, calling_prefix, country_code",
    [
        (45, 45, "DK"),
        (4566118311, 45, "DK"),
        ("+4566118311", 45, "DK"),
        ("299 80 80 80", 299, "GL"),
        (14412921234, 1441, "BM"),
    ],
)
def test_phone_country_prefix(phone_number, calling_prefix, country_code):
    assert phone_country_prefix(phone_number) == (calling_prefix, country_code)


@pytest.mark.parametrize(
    "phone_number, country_code",
    [
        (12797597235, "US"),
        (14458883022, "US"),
        (18204000053, "US"),
    ],
)
def test_us_country(phone_number, country_code):
    assert phone_country(phone_number) == country_code



import pytest

from phone_iso3166.errors import InvalidCountry
from phone_iso3166.errors import InvalidNetwork
from phone_iso3166.network import country_networks
from phone_iso3166.network import network


@pytest.mark.parametrize(
    "mcc, mnc, country_code, network_name",
    [
        (238, 1, "DK", "TDC A/S"),
        (238, 2, "DK", "Telenor Denmark"),
        (425, 6, "PS", "Ooredoo Palestine"),
    ],
)
def test_network(mcc, mnc, country_code, network_name):
    assert network(mcc, mnc) == (country_code, network_name)


def test_country_networks():
    networks = country_networks("US")
    for mcc, mnc, network_name in networks:
        assert network(mcc, mnc) == ("US", network_name)


def test_invalid_country():
    with pytest.raises(InvalidCountry):
        country_networks("XX")


def test_invalid_network():
    with pytest.raises(InvalidNetwork):
        network(0, 0)
